#include <stdio.h>
#include <limits.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#define MAX_DEG 200
#define MAX_PRIMER_LEN 100
#include "unixtimer.h"

#define CONTINUE_LINE /*   */
/*   
COMMAND FOR EXECUTION : ./a.out arg1(primer length) arg2(proximity) arg2(pcr length) 
 [OPTIONS]
e.g.  ./a.out -input input.fa -plen 8 -prox 10 -prodlen 400 
       ./a.out  -input raw.fa -plen 6 -prox 10 -prodlen 300  
 
OPTIONS:
 -plen: required length of the selected primers 

 -input: input file name 

 -prox: minimum acceptable distance from the SNP to the primer, i.e., to the end closest to the SNP 

 -prodlen: length of the sequence used for  PCR 

COMMAND FOR EXECUTION : ./a.out arg1(primer length) arg2(proximity) arg2(pcr length) [OPTIONS]
*/


   long  PRIMER;
   long PROXIMITY;   
   long PCR_LENGTH; /* make sure to add PRIMER to this to take cre of the correction
                    so that PRIMER distances is from beginning to beginning */
   
   enum {TRUE, FALSE};
   short DATA, PRIM, INTOFILE,LAZY_GREEDY,ERR,TIME;

struct SEGMENT {
  long location;
  long up_down;
  long sequence_no;
  char *seq;
};

struct SEQUENCE {
  char *up_sequence, *down_sequence;
  long up_size,down_size;
  struct SEGMENT **segment;
  long total_segment;
};

int value(struct SEGMENT *, struct SEGMENT *);
char *adjust(struct SEGMENT *); 

long pcr_no;
long total_no_candid_primers=0;
long primer_count =0;
/*long potential;*/

struct SEGMENT *allocate_segments;

/*  BEGIN - DEFINITIONS FOR PRIMER SPECIFICATION   */
#define MAX_MELT_TEMP      2000.0  /* max melting temp of the canddate */
#define MIN_MELT_TEMP     -273.0  /* max melting temp of the canddate */
#define DEFAULT_SALT_CONC     50  /* is 50.0 milimolar */
#define DEFAULT_DNA_CONC      50  /* is 50.0 nanomolar */
#define MAX_NN_TM_LENGTH 36 /* The maxium length for which to use the
                               nearest neighbor model when calculating
                               oligo Tms. */
#define MAX_GC     100  /* max gc content of the distinguisher */
#define MIN_GC     0  /* max gc content of the distinguisher */

/*  END - DEFINITIONS FOR PRIMER SPECIFICATION   */


/*  BEGIN -  VARIABLES AND FUNCTIONS FOR PRIMER PROPERTIES */   
   double gc_content;  
   double gc_and_n_content(const char *, int );
   double seqtm(const char *,  /* The sequence. */
             double,   /* DNA concentration (nanomolar). */
             double,  /* Salt concentration (millimolar). */
             int, /* The maximum sequence length for using the
                                   nearest neighbor model (as implemented
                                   in oligotm. For sequences longer
                                   than this use the "GC%" formula implemented
                                   in long_seq_tm. */
             int       /* this is the length of the sequence */
             );

   double min_candid_temp = MIN_MELT_TEMP;    /* min melting temp of distinguisher */ 
   double max_candid_temp = MAX_MELT_TEMP;    /* max melting temp of distinguisher */
   double min_candid_gc = MIN_GC;    /* min  gc content of distinguisher */
   double max_candid_gc = MAX_GC;    /* max  gc content of distinguisher */
   double candid_salt_conc = DEFAULT_SALT_CONC;
   double candid_dna_conc = DEFAULT_SALT_CONC;
/*  END -  VARIABLES  AND FUNCTIONS FOR PRIMER PROPERTIES */   

int main(int argc, char *argv[]) {
  FILE *fp, *fr;
  struct SEQUENCE sequence;
/*  long total_no_candid_primers=0; */
  struct SEGMENT **all_primers;
  
  int i;
/*  char *input =NULL; */
  extern long  PRIMER;
  extern long PROXIMITY;
  extern long PCR_LENGTH;
 
/*   Function definitions - Begin  */
   void read_a_sequence_var(FILE *, struct SEQUENCE *);
   void segment_sequence(struct SEQUENCE *);
   void Collect_primers(struct SEQUENCE *, struct SEGMENT **, long);
   void Count_primers(struct SEQUENCE *, long);
   void Load_All_Primers(FILE *,struct SEGMENT **, long);  
   void Sort_All_Primers(int (*compare)(), struct SEGMENT **, long, long);
   int Compare_Segment(struct SEGMENT *, struct SEGMENT *);
   void  Greedy_Primer_Selection(struct SEGMENT **, long, FILE*);
/*   Function definitions - End  */

/* default values */
    PRIMER=10;
    PROXIMITY=0;
    PCR_LENGTH=1000;

/*   the negative term is to correct for the fact that we count from 
    beginning to beginning of the primers*/

    DATA =TRUE;
    PRIM =FALSE;
    INTOFILE=FALSE;
    ERR=FALSE;
    TIME=TRUE;
    LAZY_GREEDY=TRUE;
    
      for(i =1; i<argc; i++)
        if(strcmp(argv[i],"-time")==0)
             TIME=TRUE;
/*
        else if( strncmp(argv[i], "-input", strlen("-input")) == 0 )
         {
          input= argv[++i];
         }
*/
        else if( strncmp(argv[i], "-prox", strlen("-prox")) == 0 )
         {
          PROXIMITY= atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-prodlen", strlen("-pcrlen")) == 0 )
         {
          PCR_LENGTH= atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-plen", strlen("-plen")) == 0 )
         {
          PRIMER= atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-max_melt_temp", strlen("-max_melt_temp")) == 0 )
         {
          max_candid_temp = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-min_melt_temp", strlen("-min_melt_temp")) == 0 )
         {
          min_candid_temp = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-min_gc", strlen("-min_gc")) == 0 )
         {
          min_candid_gc = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-max_gc", strlen("-max_gc")) == 0 )
         {
          max_candid_gc = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-salt_conc", strlen("-salt_conc")) == 0 )
         {
          candid_salt_conc = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-dna_conc", strlen("-dna_conc")) == 0 )
         {
          candid_dna_conc = atof(argv[++i]);
         }
        else
          {
             ERR=TRUE;
          }

      if(ERR==TRUE)
        {
 fprintf(stderr,"Usage:\n %s [options] < input_file [ > output_file ]\n\n", argv[0]);
         fprintf(stderr, "OPTIONS\n"); 
         fprintf(stderr,"-plen N              : primer length (default: 10)\n");
         fprintf(stderr,"-prox N              : minimum distance between primer and amplification locus (default: 0)\n");
         fprintf(stderr,"-prodlen N           : maximum product length (default: 1000)\n");
         fprintf(stderr,"-max_melt_temp N     : maximum primer temperature (default: none)\n");
         fprintf(stderr,"-min_melt_temp N     : maximum primer temperature (default: none)\n");
         fprintf(stderr,"-min_gc N            : minimum gc content of the primers (default: 0 %%)\n");
         fprintf(stderr,"-max_gc N            : maximum gc content of the primers (default: 100 %%)\n");
         fprintf(stderr,"-salt_conc N         : salt concentration in the reaction (default: 50 milimolar)\n" );
         fprintf(stderr,"-dna_conc N          : dna concentration in the reaction (default: 50 nanomolar)\n" );
         exit(0);
       }

/* 
  if((fp=fopen(input,"r"))==0)
     {
      printf("Error: Cannot open input file\n"); 
      exit(0);
    }
*/

  fp=stdin;
  fr=stdout;

  pcr_no = 0;

  while(!feof(fp)) {
     read_a_sequence_var(fp,&sequence);
     segment_sequence(&sequence);
     /*total_no_candid_primers+=sequence.total_segment; */
     Count_primers(&sequence, pcr_no);
     pcr_no++;
     free(sequence.up_sequence);
     free(sequence.down_sequence);
     free(allocate_segments);
     free(sequence.segment);
  } 


  /* allocate the  segments pointers and allocate segments to the pointer and
   also allocate space for the primer candidates */ 

  if((all_primers = (struct SEGMENT **)malloc(total_no_candid_primers*sizeof(struct SEGMENT *)))==NULL)  
    {
      fprintf(stderr, "Cannot collect memory to store the all primers at the initial stage\n");
      exit(0);
    }
  
  if(total_no_candid_primers == 0)
    {
      fprintf(stderr, "Cannot collect any feasible set of primers\n");
      exit(0);
    }

   for(i=0;i < total_no_candid_primers; i++)
    {
      if((all_primers[i]=(struct SEGMENT *)malloc(sizeof(struct SEGMENT)))==0)
        {
          printf("%d  Failing at primer no \n", i);
          printf("Error: cannot allocate memory\n");
          exit(0);
        }
      if((all_primers[i]->seq = (char *)malloc((PRIMER+1)*sizeof(char)))==0)
        {
          printf("Error: cannot allocate memory\n");
          exit(0);
        }
    }
   

  fseek(fp,0,SEEK_SET);
  pcr_no = 0;
  while(!feof(fp)) {
     read_a_sequence_var(fp,&sequence);
     segment_sequence(&sequence);
     /*total_no_candid_primers+=sequence.total_segment; */
     Collect_primers(&sequence, all_primers, pcr_no);
     pcr_no++;
     free(sequence.up_sequence);
     free(sequence.down_sequence);
     free(allocate_segments);
     free(sequence.segment);
  } 


   start_timer();

  Sort_All_Primers(Compare_Segment, all_primers, 0, total_no_candid_primers-1);

 if(DATA==TRUE)
  {
   fprintf(fr,"USER DEFINED PARAMETERS:\n");
   fprintf(fr,"PRIMER LENGTH                      : %ld\n", (long)PRIMER);
   fprintf(fr,"PRODUCT LENGTH                     : %ld\n", (long)PCR_LENGTH);
   fprintf(fr,"MIN PROXIMITY TO SNP               : %ld\n",(long)PROXIMITY);
   fprintf(fr,"NUMBER OF TARGETS                  : %ld\n", pcr_no);
   fprintf(fr,"MELTING TEMP RANGE                 : ( %f , %f )\n", min_candid_temp, max_candid_temp);
   fprintf(fr,"GC CONTENT RANGE                   : ( %f , %f )\n", min_candid_gc,max_candid_gc);
   fprintf(fr,"PRIMER SALT CONC                   : %f\n", candid_salt_conc);
   fprintf(fr,"PRIMER DNA CONC                    : %f\n\n\n", candid_dna_conc);
  }

  Greedy_Primer_Selection(all_primers, total_no_candid_primers, fr);

  fclose(fp);
  return(0);
}


void read_a_sequence_var(FILE *fp, struct SEQUENCE *add_sequence)
{
    char c;
    int i;
   long pcr_up_size=0, pcr_down_size=0; 
    

    while(!feof(fp)) 
     {
      c =fgetc(fp);
      if(c=='a' || c=='t' || c=='c' || c=='g')
        break;
     }
    fseek(fp, -1, SEEK_CUR); 

    while(!feof(fp))
    {
      c =fgetc(fp);
      pcr_up_size++;
      if(!(c=='a' || c=='t' || c=='c' || c=='g'))
        break; 
    }  
    fseek(fp, -pcr_up_size, SEEK_CUR); 
    add_sequence->up_sequence = (char *)malloc((pcr_up_size)*sizeof(char));
    pcr_up_size--;
    for(i=0; i <pcr_up_size;i++)
       add_sequence->up_sequence[i] = fgetc(fp);
    add_sequence->up_sequence[pcr_up_size] = '\0';
    add_sequence->up_size = pcr_up_size;

    while(!feof(fp))
     {
      c =fgetc(fp);
      if(c=='a' || c=='t' || c=='c' || c=='g')
        break;
     }
    fseek(fp, -1, SEEK_CUR); 

    while(!feof(fp))
    {
      c =fgetc(fp);
      pcr_down_size++;
      if(!(c=='a' || c=='t' || c=='c' || c=='g'))
        break; 
    }  

    fseek(fp, -pcr_down_size, SEEK_CUR); 
    add_sequence->down_sequence = (char *)malloc((pcr_down_size)*sizeof(char));
    pcr_down_size--;
    for(i=0; i <pcr_down_size;i++)
       add_sequence->down_sequence[i] = fgetc(fp);
    add_sequence->down_sequence[pcr_down_size] = '\0';
    add_sequence->down_size = pcr_down_size;

    while(!feof(fp))
     {
      c =fgetc(fp);
      if(c=='\n')
        break;
     }
      c =fgetc(fp);
}



 void segment_sequence(struct SEQUENCE *sequence)
 {
   extern long  PRIMER;
   extern long PROXIMITY;
   extern long PCR_LENGTH;
   long up_start, up_end; 
   long down_start, down_end; 
   long total,count;
   int i;

/* number of primers for the upstream primer */
   if(sequence->up_size > (PCR_LENGTH-PROXIMITY-PRIMER))
     up_start = sequence->up_size-(PCR_LENGTH-PROXIMITY-PRIMER);
   else
     up_start =0;
   up_end = sequence->up_size-PRIMER-PROXIMITY;
   if(up_end >= up_start) 
     total = up_end-up_start+1;
   else 
     total =0;

/* number of primers for the downstream + upstreamprimer */
   down_start = PROXIMITY;
   if(sequence->down_size > (PCR_LENGTH-PROXIMITY-PRIMER))
      down_end = PCR_LENGTH-PROXIMITY-PRIMER -PRIMER;
   else 
      down_end=sequence->down_size -PRIMER;
   if((down_end -down_start) >= 0)
     total+=down_end-down_start+1;   /* careful */
   sequence->segment = (struct SEGMENT **)malloc(total*sizeof(struct SEGMENT *));
   allocate_segments=(struct SEGMENT *)malloc(total*sizeof(struct SEGMENT));
   sequence->total_segment = total;

/* allocate the up stream primers follwed by down stream */
    count=0;
    if(up_end-up_start >=0) 
      for(i=up_start;i <= up_end;i++)
       {
         allocate_segments[count].up_down =0;
         allocate_segments[count].seq = (sequence->up_sequence+i);
         allocate_segments[count].location = sequence->up_size - i; 
         sequence->segment[count] = allocate_segments+count ;
         count++;
       }
    if(down_end -down_start >=0) 
      for(i=down_start;i <= down_end;i++)
       {
         allocate_segments[count].up_down = 1;
         allocate_segments[count].seq = (sequence->down_sequence+i);
         allocate_segments[count].location  = -i-1; 
         sequence->segment[count] = allocate_segments+count ;
         count++;
       }
}


/* this function selects the primers using a greedy approach */
void Collect_primers(struct SEQUENCE *sequence, struct SEGMENT **all_primers, long pcr_no) 
{
   int i,j;
   char Seq[1000];
   char *temp_str; 
   int len;
   double temperature, gc_content;
   long no_primers;

   no_primers = total_no_candid_primers;

   for(i=0;i < sequence->total_segment; i++)
    { 
  /* copy the candidate to a sequence to compute the temp and the gc content */
      temp_str = adjust(sequence->segment[i]);
      len =strlen(temp_str);
      for(j=0;j< len;j++)
        {
          Seq[j]=toupper(temp_str[j]);
        }
      Seq[len] ='\0';

      temperature=seqtm(Seq, candid_salt_conc, candid_dna_conc, MAX_NN_TM_LENGTH, len);
      gc_content=gc_and_n_content(Seq, len);
 if((temperature>min_candid_temp)&&(temperature<max_candid_temp)&&(gc_content>=min_candid_gc)&&(gc_content<=max_candid_gc))
        {
           temp_str=adjust(sequence->segment[i]);
           for(j=0;j<=PRIMER;j++)
              all_primers[primer_count]->seq[j] = temp_str[j];
           all_primers[primer_count]->sequence_no= pcr_no;
           all_primers[primer_count]->location= sequence->segment[i]->location;
           all_primers[primer_count]->up_down=sequence->segment[i]->up_down;
           primer_count++;
        }
    }
}

void Count_primers(struct SEQUENCE *sequence, long pcr_no) 
{
   int i,j;
   char Seq[1000];
   char *temp_str; 
   int len;
   double temperature, gc_content;

   for(i=0;i < sequence->total_segment; i++)
    { 
  /* copy the candidate to a sequence to compute the temp and the gc content */
      temp_str = adjust(sequence->segment[i]);
      len =strlen(temp_str);
      for(j=0;j< len;j++)
        {
          Seq[j]=toupper(temp_str[j]);
        }
      Seq[len] ='\0';

      temperature=seqtm(Seq, candid_salt_conc, candid_dna_conc, MAX_NN_TM_LENGTH, len);
      gc_content=gc_and_n_content(Seq, len);
 if((temperature>min_candid_temp)&&(temperature<max_candid_temp)&&(gc_content>=min_candid_gc)&&(gc_content<=max_candid_gc))
        {
         total_no_candid_primers++; 
        }
    }
}


/*  this function uses quick sort to sortthe  primers */
void Sort_All_Primers(int (*compare)(),struct SEGMENT **primers, long p, long r)
 {
   long q;
   void Sort_All_Primers(int (*compare)(),struct SEGMENT **, long, long);
   int primer_partition(int (*compare)(), struct SEGMENT **,long, long);
   if(p < r)
   {
      q = primer_partition(compare, primers,p,r);
      Sort_All_Primers(compare, primers,p,q);
      Sort_All_Primers(compare, primers,q+1,r);
   }
 } 
    
int primer_partition(int (*compare)(), struct SEGMENT **primers,long p, long r)
  { 
   long i,j;
   struct SEGMENT *x,*temp;
   x = primers[p];
   i = p-1;
   j = r+1;
   while(j > i) 
   {
     do{
       j--;
     } while(compare(primers[j],x)>0); 

     do {
       i++; 
     } while(compare(primers[i],x) < 0); 

     if(i <j)
      {
        temp=primers[i];
        primers[i] = primers[j];
        primers[j] = temp;
      }     
     else
        return(j);  }
    return(0);
 } 


int Compare_Segment(struct SEGMENT *segment1, struct SEGMENT *segment2)
 {
   if(strcmp(segment1->seq,segment2->seq)<0)
       return(-1);
   else if(strcmp(segment1->seq,segment2->seq)>0)
       return(1);
   else 
         if(segment1->up_down < segment2->up_down)
           return(-1);
         else if(segment1->up_down > segment2->up_down)
           return(1);
         if(segment1->sequence_no < segment2->sequence_no)
           return(-1);
         else if(segment1->sequence_no > segment2->sequence_no)
           return(1);
         else 
           return(0);
 }

/* this function returns a string to compare */
static char temp[MAX_PRIMER_LEN+1];

char *adjust(struct SEGMENT *segmen) 
  {
     long i;
     for(i=0;i <PRIMER;i++)
       {
        if(segmen->up_down == 0) /* changed from 1*/
           temp[i]=segmen->seq[i]; 
        else
          if(segmen->seq[i]=='a')
           temp[PRIMER-1-i]='t';
          else if(segmen->seq[i]=='t')
           temp[PRIMER-1-i]='a';
          else if(segmen->seq[i]=='c')
           temp[PRIMER-1-i]='g';
          else if(segmen->seq[i]=='g')
           temp[PRIMER-1-i]='c';
          else
          {
            printf( "Unexpected letter in adjust\n" );
            exit(0);
          }
       }

     temp[PRIMER] = '\0';   
     return(temp); 
  }

struct RECORD 
 {
   long sequence_no;
   long location;
   long up_down; 
   long active;
 };
struct PRIMER_LIST
{
   long status;
   char *String;
   long  degree,old_degree;
   struct RECORD *records;
   long Pot_Diff;
};
  


void Greedy_Primer_Selection(struct SEGMENT **all_primers, long total_no_primers,  FILE *fr)
{
   void Extract_Primers(struct PRIMER_LIST *, long, FILE *);
   int i,j,k,similar;
   long  no_primer_nodes,no_new_nodes;
   long start, end;
   struct PRIMER_LIST *primers;
   struct PRIMER_LIST *new_primers;

   void Ret_Common_Seg(long *,long *,struct SEGMENT **,long);
   void Load_PRIMER_LIST(long,long,struct SEGMENT **,long,struct PRIMER_LIST *,long);

   if(total_no_primers <2)
   {
      printf("Error: Too few primers\n");
      exit(0);
   } 
   no_primer_nodes=1;
   for(i=0;i< total_no_primers-1;i++) 
   {
     if(!(strcmp(all_primers[i]->seq, all_primers[i+1]->seq) ==0 && all_primers[i]->up_down 
   CONTINUE_LINE == all_primers[i+1]->up_down))
           no_primer_nodes++;
   }
   
/* load all all the distinct primers to the data structures PRIMER_LIST */
   primers = (struct PRIMER_LIST *)malloc(no_primer_nodes*sizeof(struct PRIMER_LIST));

   do {
      Ret_Common_Seg(&start, &end, all_primers, total_no_primers);
      if(start != -1)
         Load_PRIMER_LIST(start,end,all_primers, total_no_primers, primers, no_primer_nodes);
   } while(start >=0);


/*   Merge primers from both sides to one node */
   similar=0;
   for(i=0;i< no_primer_nodes-1;i++)
      if(strcmp(primers[i].String,primers[i+1].String)==0)
        similar++;
   no_new_nodes = no_primer_nodes - similar;
   new_primers = (struct PRIMER_LIST *)malloc(no_new_nodes*sizeof(struct PRIMER_LIST));

   j=0;
   for(i=0;i< no_primer_nodes;i++)
     {
      
      if(i< no_primer_nodes-1) 
        {
         if(strcmp(primers[i].String,primers[i+1].String)==0)
           new_primers[j].records=(struct RECORD *)malloc((primers[i].degree+primers[i+1].degree)*sizeof(struct RECORD));
         else
           new_primers[j].records=(struct RECORD *)malloc((primers[i].degree)*sizeof(struct RECORD));
        }
      else
         new_primers[j].records=(struct RECORD *)malloc((primers[i].degree)*sizeof(struct RECORD));

      new_primers[j].String=primers[i].String;
      new_primers[j].degree = primers[i].degree;
      new_primers[j].old_degree = primers[i].old_degree;
      for(k = 0; k < primers[i].degree; k++)
        {
         new_primers[j].records[k].sequence_no = primers[i].records[k].sequence_no;
         new_primers[j].records[k].location = primers[i].records[k].location;
         new_primers[j].records[k].up_down = primers[i].records[k].up_down;
        }
      if(i >= no_primer_nodes-1) break;

      if(strcmp(primers[i].String,primers[i+1].String)==0)
       {
         new_primers[j].degree = primers[i].degree + primers[i+1].degree;
         new_primers[j].old_degree = primers[i].old_degree + primers[i+1].old_degree;
         for(k = primers[i].degree;k< primers[i].degree+primers[i+1].degree; k++)

           {
             new_primers[j].records[k].sequence_no = primers[i+1].records[k-primers[i].degree].sequence_no;
             new_primers[j].records[k].location = primers[i+1].records[k-primers[i].degree ].location;
             new_primers[j].records[k].up_down = primers[i+1].records[k-primers[i].degree].up_down;
           }
         i++;
       }
       j++;
     }

   Extract_Primers(new_primers,  no_new_nodes, fr);
   for(i=0;i< no_primer_nodes;i++)
    if(primers[i].records!=NULL) 
     free(primers[i].records);
   if(primers!=NULL);
      free(primers);
  }


void Extract_Primers(struct PRIMER_LIST *primers, long  no_primers, FILE *fr)
 {
  long i, j ;
  long no_seq_covered=0;
  long *seq_fill[2];
  long *P_sel, P_sel_no, *P_active;
  long Potential_Fun(struct PRIMER_LIST *, long ,long,long, long *, long *);
  extern long pcr_no;
  long Chi_diff_max,Chi_1,Chi_2;
  long max_primer;
  long Max_Delta=0;
  long actual;
  int new_flag;
  long *left, *right;
   
  long len;
  double temperature, gc_content;
  char Seq[1000];

  P_sel_no=0;
  P_sel = (long *)malloc(4*pcr_no*sizeof(long));
  P_active = (long *)malloc(4*pcr_no*sizeof(long));
  for(i=0;i <4*pcr_no; i++)
   {
    P_sel[i] = 0;
    P_active[i]=1;
   }

  seq_fill[0] = (long *)malloc(pcr_no*sizeof(long));
  seq_fill[1] = (long *)malloc(pcr_no*sizeof(long));

  for(i=0;i < pcr_no; i++)
   {
    seq_fill[0][i] = 0;
    seq_fill[1][i] = 0;
   }

/* set the status variable to 0 at the beginning */
   for(i =0;i<no_primers; i++)
    {
     primers[i].status = 1; 
     primers[i].Pot_Diff = LONG_MAX;
    }


    if((left=(long *)malloc(pcr_no*sizeof(long)))==0)
     {
       printf("Cannot allocate memory\n");
       exit(0);
     }
    if((right=(long *)malloc(pcr_no*sizeof(long)))==0)
     {
       printf("Cannot allocate memory\n");
       exit(0);
     }
    actual =0;
    Chi_1=Potential_Fun(primers,no_primers,actual,0,left,right);
    Chi_diff_max =0;
    do 
     {

     /* choose c_0   */
       new_flag=0;
       if(Chi_diff_max>=Max_Delta)
         Max_Delta=Chi_diff_max;
       Chi_diff_max =0;

       max_primer= -1;
       for(i =0;i<no_primers; i++)
        if((primers[i].Pot_Diff >= Chi_diff_max) || LAZY_GREEDY==FALSE) 
        {
         if(primers[i].status==1) 
           {  
                 actual=0;
                 Chi_2=Potential_Fun(primers,no_primers,actual,i,left,right);

                if(Chi_1< Chi_2)
                 {
                  printf("Chi_1  = %ld   Chi_2 = %ld\n",Chi_1,Chi_2);
                  printf("This is a mistake\n");
                  exit(0);
                 }
                primers[i].Pot_Diff=Chi_1-Chi_2; 
                if(Chi_diff_max <=primers[i].Pot_Diff) 
                  {
                   max_primer=i;
                   Chi_diff_max =primers[i].Pot_Diff; 
                  }
           }
        } /* end for the for */
     /* choose c_0   */


     /* C <- c \{c0}   and P <- P u {c0}  statements */
      if(max_primer >= 0  && Chi_diff_max !=0)
       {
         actual=1;
         Chi_1=Potential_Fun(primers,no_primers,actual,max_primer,left,right);

         P_sel[P_sel_no]=max_primer; 
         P_sel_no++;
         primers[max_primer].status = 2;

         for(i =0;i<primers[max_primer].degree; i++)
            seq_fill[primers[max_primer].records[i].up_down][primers[max_primer].records[i].sequence_no]=1;
       }
     }while(Chi_diff_max !=0);   


/* count the number of seqences covered */

    no_seq_covered=0;
     for(i =0;i<pcr_no;i++)
       if((seq_fill[0][i]==1) && (seq_fill[1][i]==1) && right[i]-left[i] <= PCR_LENGTH)  
          no_seq_covered++;

  if(TIME==TRUE)
     fprintf(fr, "CPU SECONDS                        : %.2lf\n", cpu_seconds() );

/*  fprintf(fr,"Log(Max Entropy) Change= %lf\n",logf((float)Max_Delta)); */
  fprintf(fr,"NUMBER OF PRIMERS SELECTED         : %ld\n",P_sel_no);
  fprintf(fr,"NUMBER OF TARGETS COVERED          : %ld\n\n",no_seq_covered);


/* list out the primers   */
  for(i=0;i<PRIMER-6; i++)
    fprintf(fr," ");
  fprintf(fr,"PRIMER(5'-3')       MELT  TEMP            GC\n");

  for(i =0;i<no_primers; i++)
      if(primers[i].status==2) 
       {
         len =strlen(primers[i].String);
         for(j=0;j< len;j++)
           {
             Seq[j]=toupper(primers[i].String[j]);
           }
         Seq[len] ='\0';

         temperature=seqtm(Seq, candid_salt_conc, candid_dna_conc, MAX_NN_TM_LENGTH, len);
         gc_content=gc_and_n_content(Seq, len);

         fprintf(fr,"   %s         %6.2lf            %6.2lf\n",primers[i].String, temperature, gc_content);

       }
 
/* free the allocated memory  */
   free(seq_fill[0]);
   free(seq_fill[1]);
}


long Potential_Fun(struct PRIMER_LIST *primers, long  no_primers, long actual_change,long n,long *left,long *right)
  {
     extern long PCR_LENGTH;
     extern long PRIMER;
     static long *storeleft;
     static long *storeright;
     static int *flag;
     static long first=0;
     long i;
     extern long pcr_no;
   
     long L, L_Prime;
     long small_L;
     static long *Phi;
     static long *storePhi;
     long static Total_Phi;  /* it is the total potential */
     long static storeTotal_Phi;  /* it is the old previous total potential */
     long  TempTotal_Phi;  /* it is the total potential */


/* initialise L and P_Primer */
     L = PCR_LENGTH;
     small_L=PRIMER; 

     L_Prime = L-2*small_L+1;
     
   /* ONLY   FIRST TIME INITIALIZATION and malloc */
    if(first ==0)
     {
       if((storeleft=(long *)malloc(pcr_no*sizeof(long)))==0)
       {
         printf("Cannot allocate memory\n");
         exit(0);
        }
       if((storeright=(long *)malloc(pcr_no*sizeof(long)))==0)
        {
         printf("Cannot allocate memory\n");
         exit(0);
        }
       if((Phi=(long *)malloc(pcr_no*sizeof(long)))==0)
        {
         printf("Cannot allocate memory\n");
         exit(0);
        }
       if((storePhi=(long *)malloc(pcr_no*sizeof(long)))==0)
        {
         printf("Cannot allocate memory\n");
         exit(0);
        }
       if((flag=(int *)malloc(pcr_no*sizeof(int)))==0)
        {
         printf("Cannot allocate memory\n");
         exit(0);
        }

/* intialization of the left and right variables  */
/* also set the flag to 0  */
        for(i =0;i < pcr_no;i++)
         {
          left[i] = -L_Prime-1;
          right[i] = L_Prime+small_L;
          storeleft[i]=left[i];
          storeright[i]=right[i];
          flag[i]=0;
         }
        first++;
        for(i =0;i < pcr_no;i++)
         {
           Phi[i] = right[i]-left[i] -small_L -1;
           storePhi[i] = Phi[i];
         }
        Total_Phi = 0;
        storeTotal_Phi = 0;
        for(i =0;i < pcr_no;i++)
         {
           Total_Phi+=Phi[i];
           storeTotal_Phi+=Phi[i];
         }
        return(Total_Phi);
     } /* end if first =0 */

/* first save the current left and right values  */
      for(i =0; i< primers[n].degree; i++)
        {
         storeleft[primers[n].records[i].sequence_no]=left[primers[n].records[i].sequence_no];
         storeright[primers[n].records[i].sequence_no]=right[primers[n].records[i].sequence_no];
       }

/*   now update the new locations for the primer number n */
     for(i =0; i< primers[n].degree; i++)
        if(primers[n].records[i].up_down==1)  /* if candidate is from the  left side */
          {
          if(left[primers[n].records[i].sequence_no]<primers[n].records[i].location)
             left[primers[n].records[i].sequence_no]=primers[n].records[i].location;
          }
        else /* if candidate is from the  right side */
          {
          if(right[primers[n].records[i].sequence_no]>primers[n].records[i].location)
             right[primers[n].records[i].sequence_no]=primers[n].records[i].location;
          }



  /* first store the old Phi s  */
  for(i =0; i< primers[n].degree; i++)
     storePhi[primers[n].records[i].sequence_no]=Phi[primers[n].records[i].sequence_no];

/* calculate the new potential for every sequence */
    for(i =0;i < primers[n].degree ;i++)
      Phi[primers[n].records[i].sequence_no]=right[primers[n].records[i].sequence_no]-left[primers[n].records[i].sequence_no]-small_L-1;

/* compute the Total_Phi which is the total  potential */
 /* substract the old potentials for the sequences relevant to the primer */
 /*  and then add the new potentials for the sequences  */
     for(i =0;i < primers[n].degree;i++)
      {
       if(flag[primers[n].records[i].sequence_no]==0)
         {
          if(L_Prime-1 >= Phi[primers[n].records[i].sequence_no])
            Total_Phi+=L_Prime-1;
          else
            Total_Phi+=Phi[primers[n].records[i].sequence_no];

          if(L_Prime-1 >= storePhi[primers[n].records[i].sequence_no])
            Total_Phi-=L_Prime-1;
          else
            Total_Phi-=storePhi[primers[n].records[i].sequence_no];
  
          flag[primers[n].records[i].sequence_no]=1; /* set the flag */
         }
      }
 /* unset the flags  */
     for(i =0;i < primers[n].degree;i++)
       flag[primers[n].records[i].sequence_no]=0;

    if(actual_change==0) /* the primer is not yet included but just tested then retain old value */
     {
      for(i =0; i< primers[n].degree; i++)
        {
         left[primers[n].records[i].sequence_no]=storeleft[primers[n].records[i].sequence_no];
         right[primers[n].records[i].sequence_no]=storeright[primers[n].records[i].sequence_no];
       }
      }
    else
     {
      for(i =0; i< primers[n].degree; i++)
        {
         storeleft[primers[n].records[i].sequence_no]=left[primers[n].records[i].sequence_no];
         storeright[primers[n].records[i].sequence_no]=right[primers[n].records[i].sequence_no];
       }
     }

    if(actual_change==0) /* the primer is not yet included but just tested then retain old value */
     {
      for(i =0; i< primers[n].degree; i++)
         Phi[primers[n].records[i].sequence_no]=storePhi[primers[n].records[i].sequence_no];
     }
    else
     {
      for(i =0; i< primers[n].degree; i++)
         storePhi[primers[n].records[i].sequence_no]=Phi[primers[n].records[i].sequence_no];
     }

    TempTotal_Phi=Total_Phi;
    if(actual_change==0) /* the primer is not yet included but just tested then retain old value */
         Total_Phi=storeTotal_Phi;
    else
         storeTotal_Phi=Total_Phi;

     return(TempTotal_Phi);

  }

void Load_PRIMER_LIST(start, end,all_primers,total_no_primers,primers, no_primer_nodes)
long start, end;
struct SEGMENT **all_primers;
long total_no_primers;
struct PRIMER_LIST *primers;
long no_primer_nodes;
 {
  long i,degree;
  long sequen_no ;
  long max_min =0,max_min_index;
  long fill;
  static long upto =0;
  if(start == -1) return;

  degree=1;
  sequen_no = all_primers[start]->sequence_no;
  for(i=start; i<=end;i++)
   {
   if(all_primers[i]->sequence_no != sequen_no) 
     {
        sequen_no=all_primers[i]->sequence_no; 
        degree++;
     }
   }


  primers[upto].degree = degree;
  primers[upto].old_degree = degree; 
  primers[upto].String = all_primers[start]->seq;
  if(degree == 0) { printf("There is at least one bug\n"); exit(0);}
  if((primers[upto].records = (struct RECORD *)malloc(degree*sizeof(struct RECORD)))==0)
   {
     printf("Error: cannot allocate memory\n");
     exit(0);
   } 
  fill =0;
  if(start ==end)
    {
      primers[upto].records[fill].sequence_no=all_primers[start]->sequence_no;
      primers[upto].records[fill].up_down=all_primers[start]->up_down;
      primers[upto].records[fill].location=all_primers[start]->location;
      fill++;
      upto++;
      return;
    }
  

  max_min =all_primers[start]->location;
  sequen_no = all_primers[start]->sequence_no;
  max_min_index = start;

  for(i=start; i<=end;i++)
    {

      if(max_min_index == -1)
        {
         max_min =all_primers[i]->location;
         max_min_index = i;
        }

      if(all_primers[start]->up_down ==1)
       {
         if(max_min < all_primers[i]->location)
          {
            max_min = all_primers[i]->location;
            max_min_index=i;
          }
       }
     else
       {
         if(max_min > all_primers[i]->location)
          {
            max_min = all_primers[i]->location;
            max_min_index=i;
          }
       }


      if(i == end)
      {
         primers[upto].records[fill].up_down=all_primers[i]->up_down;
         primers[upto].records[fill].location=all_primers[max_min_index]->location;
         primers[upto].records[fill].sequence_no = all_primers[i]->sequence_no;
         fill++;
      }

    if(i< end)
    {
      if(all_primers[i]->sequence_no != all_primers[i+1]->sequence_no)
      {
         primers[upto].records[fill].up_down=all_primers[i]->up_down;
         primers[upto].records[fill].location=all_primers[max_min_index]->location;
         primers[upto].records[fill].sequence_no= (all_primers[i]->sequence_no);
         fill++;
         max_min =all_primers[i+1]->location;
         max_min_index = -1 ;
      } 
     } /* fi */
    }
    upto++;
 }


void Ret_Common_Seg( start, end,  all_primers, total_no_primers)
long *start, *end;
struct SEGMENT **all_primers;
long total_no_primers;
 {
   static long current =0;

   if(current == total_no_primers)
    {
      *start =-1; 
      return; 
    }
 
   *start = current;

   if(current == total_no_primers-1)
   {
     *end = current;
     current++;
     return;
   }
 
   while(strcmp(all_primers[current]->seq, all_primers[current+1]->seq) ==0
    CONTINUE_LINE && all_primers[current]->up_down == all_primers[current+1]->up_down)
   {
      current++;     
      if(current == total_no_primers-1)
      {
       *end = current;
       current++;
       return;
      }
   }
   *end = current;
   current++;
   return;
 }

